"""Redox potential workflow."""

from typing import Annotated, Any, TypeVar

from pydantic import AfterValidator, ValidationInfo, field_validator, model_validator

from ..base import round_optional_float
from ..mode import Mode
from ..solvent import Solvent
from ..types import UUID
from .multistage_opt import MultiStageOptMixin
from .workflow import MoleculeWorkflow

_T = TypeVar("_T")


class RedoxPotentialWorkflow(MoleculeWorkflow, MultiStageOptMixin):
    """
    Workflow for computing spin states of molecules.

    Uses the modes from MultiStageOptSettings.

    Inherited
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow
    :param multistage_opt_settings: set by mode unless mode=MANUAL (ignores additional settings if set)
    :param xtb_preopt: pre-optimize with xtb (sets based on mode when None)
    :param constraints: constraints to add
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    Overridden:
    :param mso_mode: Mode for MultiStageOptSettings
    :param solvent: solvent to use for optimization

    New:
    :param reduction: whether or not to calculate the reduction half-reaction
    :param oxidation: whether or not to calculate the oxidation half-reaction
    :param neutral_molecule: UUID of the calculation for the neutral molecule
    :param anion_molecule: UUID of the calculation for the anion molecule
    :param cation_molecule: UUID of the calculation for the cation molecule
    :param reduction_potential: the final potential, in V
    :param oxidation_potential: the final potential, in V

    Legacy:
    :param redox_type: one of "reduction" or "oxidation"
    :param redox_potential: the corresponding potential, in V
    """

    solvent: Solvent = Solvent.ACETONITRILE

    reduction: bool = True
    oxidation: bool = True

    # legacy values - remove in future release!
    redox_type: str | None = None
    redox_potential: float | None = None

    # UUIDs
    neutral_molecule: UUID | None = None
    anion_molecule: UUID | None = None
    cation_molecule: UUID | None = None

    reduction_potential: Annotated[float | None, AfterValidator(round_optional_float(6))] = None
    oxidation_potential: Annotated[float | None, AfterValidator(round_optional_float(6))] = None

    @field_validator("solvent", mode="before")
    @classmethod
    def only_mecn_please(cls, val: Solvent | None) -> Solvent:
        """Only MeCN please!"""
        if val != Solvent.ACETONITRILE:
            raise ValueError("Only acetonitrile permitted!")

        return val

    @field_validator("constraints", "transition_state")
    @classmethod
    def turned_off(cls, value: _T, info: ValidationInfo) -> _T:
        if value:
            raise ValueError(f"{info.field_name} not supported in redox potential workflows.")

        return value

    @model_validator(mode="before")
    @classmethod
    def set_mode_and_mso_mode(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Set the MultiStageOptSettings mode to match current redox potential mode, and select mode if `Auto`."""
        if ("mode" not in values) or (values["mode"] == Mode.AUTO):
            values["mode"] = Mode.RAPID

        values["mso_mode"] = values["mode"]
        return values

    def model_post_init(self, __context: Any, /) -> None:
        """Keep back-compatible with old schema."""
        if self.redox_type == "oxidation":
            self.oxidation = True
            self.reduction = False
            if self.oxidation_potential is None:
                self.oxidation_potential = self.redox_potential
        elif self.redox_type == "reduction":
            self.oxidation = False
            self.reduction = True
            if self.reduction_potential is None:
                self.reduction_potential = self.redox_potential



from functools import partial

from pydantic import ValidationError
from pytest import raises

from stjames import Atom, Molecule


def test_molecule_pbc() -> None:
    # this is a funny joke, because it makes manganese
    glomar_explorer = partial(Molecule, charge=0, multiplicity=2, atoms=[Atom(atomic_number=25, position=[0, 0, 0])])

    mol_nopbc = glomar_explorer()
    assert mol_nopbc.cell is None

    valid_input = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))
    mol = glomar_explorer(cell={"lattice_vectors": valid_input})
    assert mol.cell is not None
    assert mol.cell.lattice_vectors == valid_input

    mol2 = glomar_explorer(cell={"lattice_vectors": valid_input, "is_periodic": (True, False, True)})
    assert mol2.cell is not None
    assert mol2.cell.lattice_vectors == valid_input
    assert mol2.cell.is_periodic == (True, False, True)

    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": valid_input, "is_periodic": (False, False, False)})

    invalid_input = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input})

    invalid_input2 = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, "invalid"))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input2})

    invalid_input3 = ((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0), (1.0, 1.0, 1.0))
    with raises(ValidationError):
        glomar_explorer(cell={"lattice_vectors": invalid_input3})


def test_xyz() -> None:
    mol = Molecule.from_xyz_lines(
        """3
H2O Molecule; smiles: O; nonsense: data: more nonsense
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )
    assert mol.smiles == "O"
    assert getattr(mol, "nonsense", None) is None

    mol = Molecule.from_xyz_lines(
        """3
name: H2O Molecule; smiles: O
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )

    assert getattr(mol, "name", None) is None

    mol = Molecule.from_xyz_lines(
        """3
cell: [(1, 2, 3), [4, 5, 6e-1], (7.0, 8.0, 9.0)]; smiles: O; is_periodic: [True, False, True]
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )

    assert mol.cell is not None
    assert mol.cell.lattice_vectors == ((1.0, 2.0, 3.0), (4.0, 5.0, 0.6), (7.0, 8.0, 9.0))
    assert mol.cell.is_periodic == (True, False, True)
    assert mol.smiles == "O"

    mol = Molecule.from_xyz_lines(
        """3
cell: [(a, b, c), [4, 5, 6e-1], (7.0, 8.0, 9.0)]; smiles: O; is_periodic: [True, False, True]
O        0.000000     0.000000     0.000000
H        0.000000    -0.757000     0.587000
H        0.000000     0.757000     0.587000
""".splitlines()
    )

    assert mol.cell is None
    assert mol.smiles == "O"
    assert getattr(mol.cell, "is_periodic", None) is None



"""Read elemental data from files."""

import json
from importlib import resources
from typing import NamedTuple

data_dir = resources.files("stjames").joinpath("data")

with data_dir.joinpath("symbol_element.json").open() as f:
    SYMBOL_ELEMENT: dict[str, int] = json.loads(f.read())

ELEMENT_SYMBOL = {v: k for k, v in SYMBOL_ELEMENT.items()}

Isotope = NamedTuple("Isotope", [("relative_atomic_mass", float), ("isotopic_composition", float), ("standard_atomic_weight", float)])
with data_dir.joinpath("nist_isotopes.json").open() as f:
    d = json.loads(f.read())

    ISOTOPES: dict[int, dict[int, Isotope]] = {
        int(k): {
            int(kk): Isotope(*vv)
            for kk, vv in v.items()  # stay open
        }
        for k, v in d.items()
    }

with data_dir.joinpath("bragg_radii.json").open() as f:
    BRAGG_RADII: dict[int, float] = json.loads(f.read())



"""pKa workflow."""

from typing import Annotated, Optional, Self

from pydantic import AfterValidator, model_validator

from ..base import Base, round_float, round_optional_float
from ..types import round_list
from .workflow import SMILESWorkflow


class MacropKaMicrostate(Base):
    """
    A microstate for pKa calculations.

    :param smiles: SMILES string for this conformer
    :param energy: free energy of this conformer
    :param charge: the total charge
    """

    smiles: str
    energy: Annotated[float, AfterValidator(round_float(3))]  # free energy
    charge: int


class MacropKaValue(Base):
    """
    Represents a change in pKa.

    :param initial_charge: the charge of the initial state
    :param final_charge: the charge of the final state
    :param pKa: the pKa for the transition
    """

    initial_charge: int
    final_charge: int
    pKa: Annotated[float, AfterValidator(round_float(3))]


class MacropKaWorkflow(SMILESWorkflow):
    """
    Workflow for calculating pKa.

    Inherited:
    :param initial_smiles:

    New:
    :param min_pH: for precomputed microstate weights, logD, etc
    :param max_pH: for precomputed microstate weights, logD, etc
    :param max_charge: max charge to consider for microstates
    :param min_charge: min charge to consider for microstates
    :param compute_aqueous_solubility: whether or not to compute aqueous solubility
    :param compute_solvation_energy: whether to run a csearch + compute the solvation energy (for Kpuu)

    Results:
    :param microstates: microstates
    :param pKa_values: macroscopic pKa values
    :param isoelectric_point: the isoelectric point (in pH units)
    :param solvation_energy: the solvation energy, in kcal/mol
    :param microstate_weights_by_pH: the % of different microstates by pH
    :param logD_by_pH: the distribution constant (water/octanol) by pH
    :param aqueous_solubility_by_pH: the log(S)/L of the compound in water, by pH
    :param kpuu_probability: the probability that Kpuu >= 0.3, the Schrodinger-determined threshold
    """

    min_pH: Annotated[float, AfterValidator(round_float(3))] = 0.0
    max_pH: Annotated[float, AfterValidator(round_float(3))] = 14.0
    max_charge: int = 2
    min_charge: int = -2
    compute_aqueous_solubility: bool = False
    compute_solvation_energy: bool = True

    microstates: list[MacropKaMicrostate] = []
    pKa_values: list[MacropKaValue] = []
    isoelectric_point: Annotated[Optional[float], AfterValidator(round_optional_float(3))] = None
    solvation_energy: Annotated[Optional[float], AfterValidator(round_optional_float(3))] = None
    kpuu_probability: Annotated[Optional[float], AfterValidator(round_optional_float(3))] = None

    microstate_weights_by_pH: list[
        tuple[
            Annotated[float, AfterValidator(round_float(3))],
            Annotated[list[float], AfterValidator(round_list(6))],
        ]
    ] = []

    logD_by_pH: list[
        tuple[
            Annotated[float, AfterValidator(round_float(3))],
            Annotated[float, AfterValidator(round_float(3))],
        ]
    ] = []

    aqueous_solubility_by_pH: list[
        tuple[
            Annotated[float, AfterValidator(round_float(3))],
            Annotated[float, AfterValidator(round_float(3))],
        ]
    ] = []

    @model_validator(mode="after")
    def check_weights(self) -> Self:
        for _, weights in self.microstate_weights_by_pH:
            if len(weights) != len(self.microstates):
                raise ValueError("Length of microstate weights doesn't match!")

        return self

    @model_validator(mode="after")
    def check_minmax_charges(self) -> Self:
        if self.min_charge >= self.max_charge:
            raise ValueError("Incoherent min/max charge specification")

        return self



from .mmcif import mmcif_dict_to_data_dict, mmcif_string_to_mmcif_dict
from .pdb import pdb_dict_to_data_dict, pdb_string_to_pdb_dict
from .utilities import fetch, open

__all__ = ["fetch", "mmcif_dict_to_data_dict", "mmcif_string_to_mmcif_dict", "open", "pdb_dict_to_data_dict", "pdb_string_to_pdb_dict"]



"""Workflows covering RBFE graph construction and endpoint FEP execution."""

from typing import Annotated, Any, Literal, Self

from pydantic import AfterValidator, PositiveInt, model_validator

from ..base import Base, round_float, round_optional_float
from ..message import Message
from ..method import Method
from ..molecule import Molecule
from ..pdb import PDB
from ..types import UUID
from .workflow import ProteinStructureWorkflow, Workflow


class TMDRBFESettings(Base):
    """
    TMD-specific simulation parameters shared across RBFE graph edges.

    This settings bundle targets the TMD runner; other engines should define
    their own settings model if their controls differ.

    :param forcefield: Registered St. James method corresponding to a force field.
    :param n_eq_steps: Equilibration steps per lambda window.
    :param n_frames: Production frames saved per lambda window.
    :param steps_per_frame: MD integration steps per saved frame.
    :param n_windows: Maximum number of lambda windows considered for bisection.
    :param min_overlap: Minimum acceptable overlap during schedule bisection.
    :param target_overlap: Desired overlap after HREX optimization.
    :param water_sampling_padding: Extra nanometers added to the solvent sampling radius.
    :param rest_max_temperature_scale: Maximum effective temperature scaling for REST.
    :param rest_temperature_scale_interpolation: Functional form used for REST scaling.
    :param local_md_steps: Number of local MD steps per frame (`0` disables local MD).
    :param local_md_k: Spring constant used during local MD.
    :param local_md_radius: Sphere radius in nanometers for the local MD region.
    :param local_md_free_reference: Whether to free the reference frame during local MD.
    :param legs: Which thermodynamic cycle legs to run (default: solvent and complex).
    """

    forcefield: Method = Method.SMIRNOFF_2_0_0_AMBER_AM1BCC
    n_eq_steps: PositiveInt = 200_000
    n_frames: PositiveInt = 2_000
    steps_per_frame: PositiveInt = 400
    n_windows: PositiveInt = 48
    min_overlap: float = 0.667
    target_overlap: float = 0.667
    water_sampling_padding: float = 0.4
    rest_max_temperature_scale: float = 1.0
    rest_temperature_scale_interpolation: Literal["exponential", "linear"] = "exponential"
    local_md_steps: int = 390
    local_md_k: float = 10_000.0
    local_md_radius: float = 1.2
    local_md_free_reference: bool = False
    legs: list[Literal["vacuum", "solvent", "complex"]] = ["solvent", "complex"]


class RBFEResult(Base):
    """
    Aggregate RBFE outcome for a single ligand.

    :param dg: Predicted binding free energy difference (kcal/mol).
    :param dg_err: Uncertainty estimate on `dg`.
    """

    dg: Annotated[float, AfterValidator(round_float(3))]
    dg_err: Annotated[float, AfterValidator(round_float(3))]


class RBFEGraphEdge(Base):
    """
    RBFE Edge definition with optional FEP edge results.

    :param mol_a: Source ligand identifier.
    :param mol_b: Target ligand identifier.
    :param core: Atom-mapping pairs describing the shared core, if any.
    :param score: Optional score used during graph construction.
    :param complex_dg: Predicted complex-leg free energy difference (kcal/mol).
    :param complex_dg_err: Uncertainty on `complex_dg`.
    :param solvent_dg: Predicted solvent-leg free energy difference (kcal/mol).
    :param solvent_dg_err: Uncertainty on `solvent_dg`.
    :param vacuum_dg: Predicted vacuum-leg free energy difference (kcal/mol).
    :param vacuum_dg_err: Uncertainty on `vacuum_dg`.
    :param ddg: Combined cycle result derived from complex and solvent legs.
    :param ddg_err: Uncertainty on `ddg`.
    :param failed: Whether a required leg failed, making ddG impossible to compute.
    """

    ligand_a: str
    ligand_b: str
    core: list[tuple[int, int]] | None = None
    score: float | None = None

    complex_dg: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    complex_dg_err: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    solvent_dg: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    solvent_dg_err: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    vacuum_dg: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    vacuum_dg_err: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    ddg: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    ddg_err: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    failed: bool = False


class RBFEGraph(Base):
    """
    Minimal RBFE graph container.

    :param edges: Directed edges describing the ligand pairs to simulate.
    """

    edges: list[RBFEGraphEdge]


class RBFEGraphWorkflow(Workflow):
    """
    Workflow that builds an rbfe graph for a set of ligands.

    :param ligands: Mapping from ligand identifiers to `Molecule` objects.
    :param graph: Optional RBFE graph output populated after the build step.
    :param mode: Graph construction strategy (`"greedy"` or `"star_map"`).
    :param hub_compound_id: Ligand identifier to serve as the hub when `mode="star_map"`.
    :param refine_cutoff: Optional cutoff used to re-run atom mapping refinement.
    :param greedy_scoring: Edge scoring heuristic for greedy mode.
    :param greedy_k_min_cut: Target edge-connectivity (`k`) for greedy augmentation.
    """

    ligands: dict[str, Molecule]
    graph: RBFEGraph | None = None

    mode: Literal["greedy", "star_map"] = "greedy"
    hub_compound_id: str | None = None
    refine_cutoff: float | None = None
    greedy_scoring: Literal["best", "jaccard", "dummy_atoms"] = "best"
    greedy_k_min_cut: PositiveInt = 3

    @model_validator(mode="after")
    def validate_builder(self) -> Self:
        """
        Validate the builder inputs.

        :raises ValueError: If star-map mode omits `hub_compound_id` or fewer than two ligands are provided.
        """
        if self.mode == "star_map" and not self.hub_compound_id:
            raise ValueError("hub_compound_id is required when mode='star_map'")
        if len(self.ligands) < 2:
            raise ValueError("Provide at least two ligands to build an RBFE graph")
        return self


class RBFEDiagnostics(Base):
    """
    Quality-control metrics gathered during endpoint RBFE.

    :param cycle_closure_rms: RMS error across completed thermodynamic cycles.
    :param windows_completed: Count of successfully converged lambda windows.
    :param windows_failed: Count of failed lambda windows.
    :param notes: Structured messages describing noteworthy events.
    """

    cycle_closure_rms: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    windows_completed: PositiveInt | None = None
    windows_failed: PositiveInt | None = None
    notes: list[Message] = []


class RelativeBindingFreeEnergyPerturbationWorkflow(ProteinStructureWorkflow):
    """
    Workflow for running relative binding free energy perturbation simulations.

    Inherited:
    :param protein: PDB of the protein, or the UUID of the protein.

    New:
    :param ligands: Mapping from ligand identifiers to `Molecule` objects.
    :param graph: RBFE graph topology.
    :param target: PDB object or the UUID of the PDB object used for simulation. DEPRECATED.

    Results:
    :param ligand_dg_results: Optional per-ligand FEP summaries produced downstream.
    :param diagnostics: Optional aggregate QC metrics.
    :param settings: Simulation controls shared across all RBFE edges.
    """

    ligands: dict[str, Molecule]
    graph: RBFEGraph
    target: PDB | UUID

    settings: TMDRBFESettings

    ligand_dg_results: dict[str, RBFEResult] | None = None
    diagnostics: RBFEDiagnostics | None = None

    @model_validator(mode="before")
    def harmonize_target_and_protein(cls, data: Any) -> Any:  # noqa: N805
        """
        Syncs data between "target" and "protein" field.
        """
        protein = data.get("protein")
        target = data.get("target")

        if target and not protein:
            data["protein"] = target
        elif not target and protein:
            data["target"] = protein

        return data



from .base import LowercaseStrEnum


class Engine(LowercaseStrEnum):
    AIMNET2 = "aimnet2"
    MACE = "mace"
    OMOL25 = "omol25"
    ORB = "orb"
    TBLITE = "tblite"
    XTB = "xtb"
    TERACHEM = "terachem"
    GPU4PYSCF = "gpu4pyscf"
    PYSCF = "pyscf"
    PSI4 = "psi4"
    OPENFF = "openff"
    EGRET = "egret"



import ast
from typing import Annotated, Self, TypeAlias

import numpy as np
import pydantic

from .base import Base
from .types import Matrix3x3, round_matrix3x3

Bool3: TypeAlias = tuple[bool, bool, bool]


class PeriodicCell(Base):
    lattice_vectors: Annotated[Matrix3x3, pydantic.AfterValidator(round_matrix3x3(6))]
    is_periodic: Bool3 = (True, True, True)

    def __repr__(self) -> str:
        """
        Return a string representation of the PeriodicCell.

        >>> PeriodicCell(lattice_vectors=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        PeriodicCell(lattice_vectors=((1.0, 2.0, 3.0), (4.0, 5.0, 6.0), (7.0, 8.0, 9.0)), is_periodic=(True, True, True))
        """
        return f"PeriodicCell(lattice_vectors={self.lattice_vectors}, is_periodic={self.is_periodic})"

    @pydantic.field_validator("lattice_vectors")
    @classmethod
    def check_tensor_3D(cls, v: Matrix3x3) -> Matrix3x3:
        if len(v) != 3 or any(len(row) != 3 for row in v):
            raise ValueError("Cell tensor must be a 3x3 list of floats")

        return v

    @pydantic.field_validator("is_periodic")
    @classmethod
    def check_pbc(cls, v: Bool3) -> Bool3:
        if not any(v):
            raise ValueError("For periodic boundary conditions, at least one dimension must be periodic!")
        return v

    @pydantic.computed_field  # type: ignore[misc, prop-decorator, unused-ignore]
    @property
    def volume(self) -> float:
        return float(np.abs(np.linalg.det(np.array(self.lattice_vectors))))

    @classmethod
    def from_string(cls, string: str) -> Self:
        """
        Create a PeriodicCell from a string representation.

        >>> PeriodicCell.from_string("[(1, -2, 3.0), [4e1, +5, 6], (7.1, 8, 9)]")
        PeriodicCell(lattice_vectors=((1.0, -2.0, 3.0), (40.0, 5.0, 6.0), (7.1, 8.0, 9.0)), is_periodic=(True, True, True))
        >>> PeriodicCell.from_string("[(a, -2, 3.0), [4e1, +5, 6], (7.1, 8, 9)]")
        Traceback (most recent call last):
        ...
        ValueError: Could not parse cell from string: [(a, -2, 3.0), [4e1, +5, 6], (7.1, 8, 9)]
        >>> PeriodicCell.from_string("[(1, -2, 3.0), [4e1, +5, 6], (7.1, 8, 9, 10)]")
        Traceback (most recent call last):
        ...
        ValueError: Cell must be a 3x3 matrix: got [(1, -2, 3.0), [4e1, +5, 6], (7.1, 8, 9, 10)]
        >>> PeriodicCell.from_string("[('a', -2, 3.0), [4e1, +5, 6], (7.1, 8, 9)]")
        Traceback (most recent call last):
        ValueError: Cell must be a 3x3 matrix of numbers: got [('a', -2, 3.0), [4e1, +5, 6], (7.1, 8, 9)]

        """
        try:
            cell = ast.literal_eval(string)
        except (ValueError, SyntaxError) as e:
            raise ValueError(f"Could not parse cell from string: {string}") from e

        if not isinstance(cell, (list, tuple)):
            raise ValueError(f"Cell must be a list or tuple: got {string}")
        if not len(cell) == 3:
            raise ValueError(f"Cell must be a 3x3 matrix: got {string}")
        if not all(isinstance(row, (list, tuple)) for row in cell):
            raise ValueError(f"Cell must be a 3x3 matrix: got {string}")
        if not all(len(row) == 3 for row in cell):
            raise ValueError(f"Cell must be a 3x3 matrix: got {string}")
        if not all(all(isinstance(x, (int, float)) for x in row) for row in cell):
            raise ValueError(f"Cell must be a 3x3 matrix of numbers: got {string}")

        return cls(lattice_vectors=cell)



"""Basic calculation workflow."""

from typing import Self

from pydantic import model_validator

from ..base import UniqueList
from ..engine import Engine
from ..settings import Settings
from ..task import Task
from ..types import UUID
from .workflow import MoleculeWorkflow


class BasicCalculationWorkflow(MoleculeWorkflow):
    """
    Workflow for a basic calculation.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow

    New:
    :param settings: Settings for running the calculation
    :param engine: Engine to use
    :param calculation_uuid: UUID of the calculation
    """

    settings: Settings
    tasks: UniqueList[Task] = [Task.ENERGY, Task.CHARGE, Task.DIPOLE]
    calculation_uuid: UUID | None = None

    # DEPRECATED - specify in settings now
    engine: Engine = None  # type: ignore [assignment]

    @model_validator(mode="after")
    def set_engine(self) -> Self:
        """Set the calculation engine."""
        self.engine = self.engine or self.settings.engine

        return self



"""Protein-binder-design workflow."""

from enum import Enum
from typing import Annotated, TypeAlias

from pydantic import AfterValidator

from ..base import Base, LowercaseStrEnum, round_optional_float
from ..protein import ProteinSequence
from ..types import UUID
from .workflow import Workflow

ProteinUUID: TypeAlias = UUID


class BoltzGenSecondaryStructure(Base):
    """
    Represents the secondary structure assignments for a protein.

    :param id: Optional identifier for this secondary structure annotation.
    :param sheet: String encoding the residue indices comprising β-sheet structures
                  (e.g., "1,3..11" for residues 1, and 3 through 11).
    :param helix: String encoding residue indices comprising helices.
    :param loop: String encoding residue indices comprising loop or coil regions.
    """

    id: str | None = None
    sheet: str | None = None
    helix: str | None = None
    loop: str | None = None


class BoltzGenProteinEntity(Base):
    """
    Represents a protein entity, either a designed or natural sequence.

    :param id: Unique identifier for the protein.
    :param sequence: Protein sequence, may contain amino acids and numbers for designed regions.
    :param secondary_structure: Optional assigned secondary structure.
    :param cyclic: Whether the protein is cyclic (True/False). Optional.
    """

    id: str
    sequence: str  # can include amino acids as well as numbers for designed regions
    secondary_structure: BoltzGenSecondaryStructure | None = None
    # binding_types: BindingType | None = None - we may want to add this later but not used in examples.
    cyclic: bool | None = None


class BoltzGenLigandEntity(Base):
    """
    Represents a ligand entity (non-protein), such as a small molecule.

    :param id: Unique identifier for the ligand.
    :param smiles: SMILES string representation of the ligand.
    """

    id: str
    smiles: str
    # binding_types: str | None = None - we may want to add this later but not used in examples.


class BoltzGenRegionSelection(Base):
    """
    Defines a region of a protein chain by specifying its chain identifier and (optionally) residue indices.

    :param chain_id: Identifier for the protein chain (e.g., 'A', 'B', etc.).
    :param residue_indices: Residues to select, specified as a string in the format "5..7,13" or "5..15,50..".
    """

    chain_id: str | None = None
    residue_indices: str | None = None


class BoltzGenProximityRegionSelection(BoltzGenRegionSelection):
    """
    Defines a region of a protein chain based on spatial proximity to a selection of residues.

    Inherits:
        BoltzGenRegionSelection

    :param radius: Radius in angstroms (Å) used to select all residues within proximity to the specified region.
    """

    radius: int | None = None


class BoltzGenBindingType(Base):
    """
    Represents the binding interface specification for a given protein chain.

    :param chain_id: Identifier for the protein chain (e.g., 'A', 'B', etc.).
    :param binding: Residue indices or regions that are required to participate in binding
        (e.g., "5..7,13" or "all" for the whole chain).
    :param not_binding: Residue indices or regions that should explicitly not participate in binding
        (e.g., "5..7,13" or "all" for excluding the entire chain).
    """

    chain_id: str | None = None
    binding: str | None = None
    not_binding: str | None = None


class BoltzGenSecondaryStructureOptions(str, Enum):
    UNSPECIFIED = "UNSPECIFIED"
    LOOP = "LOOP"
    HELIX = "HELIX"
    SHEET = "SHEET"


class BoltzGenDesignInsertion(Base):
    """
    Represents an insertion site for protein design in a specific chain.

    :param chain_id: Identifier of the chain where the insertion occurs.
    :param residue_index: Position in the chain after which the insertion is to be made.
    :param number_of_residues: Number of residues to insert at the specified site (can be a string pattern).
    :param secondary_structure: Desired secondary structure type for the inserted residues
        ("UNSPECIFIED", "LOOP", "HELIX", or "SHEET"). Optional.
    """

    chain_id: str
    residue_index: int
    number_of_residues: str
    secondary_structure: BoltzGenSecondaryStructureOptions | None = None


class BoltzGenFileEntity(Base):
    """
    Represents a protein structure input and its associated region selection and design specifications
    for the BoltzGen binder design workflow.

    :param uuid: Unique identifier for the protein structure.
    :param include: List of regions to include in the design or analysis context.
    :param exclude: List of regions to explicitly exclude from consideration (e.g., for ignoring noisy/irrelevant regions).
    :param include_proximity: List of regions defined by spatial proximity (e.g., residues within a given radius).
    :param binding_types: List of binding type constraints or permitted interface regions.
    :param design: List of regions that are being subject to design (mutable, allowed to change).
    :param secondary_structure: List of desired or annotated secondary structure definitions for selected regions.
    :param design_insertions: List of new regions to be inserted with specified properties (e.g., insertion sites, structure preferences).
    """

    uuid: ProteinUUID
    include: list[BoltzGenRegionSelection] = []
    exclude: list[BoltzGenRegionSelection] = []
    # fuse: None  - we may want to add this later but not used in examples.
    include_proximity: list[BoltzGenProximityRegionSelection] = []
    binding_types: list[BoltzGenBindingType] = []
    # structure_groups: None - we may want to add this later but not used in examples.
    design: list[BoltzGenRegionSelection] = []
    secondary_structure: list[BoltzGenSecondaryStructure] = []
    design_insertions: list[BoltzGenDesignInsertion] = []


class BoltzGenAtomSpecification(Base):
    """
    Atom specification for a protein chain, used for applying constraints or referencing atoms.

    :param chain_id: Identifier for the protein chain (e.g., "A", "B").
    :param index: Residue index the atom belongs to (integer, 1-based).
    :param atom_name: Name of the atom (e.g., "CA", "N", "O", etc.).
    """

    chain_id: str
    index: int
    atom_name: str


class BoltzGenConstraint(Base):
    """
    Describes a covalent or spatial constraint between two specified atoms in the context of protein design.

    :param atom1: First atom in the constraint.
    :param atom2: Second atom in the constraint.
    """

    atom1: BoltzGenAtomSpecification
    atom2: BoltzGenAtomSpecification


class BoltzGenInput(Base):
    """
    Represents the primary input schema for the boltzgen application.

    :param protein_entities: Protein chains that are designed or targeted for binding.
    :param ligand_entities: Small molecules or other non-protein ligands relevant to the design.
    :param file_entities: 3d protein structures and input settings related to them.
    :param constraints: Covalent bond constraints
    """

    protein_entities: list[BoltzGenProteinEntity] = []
    ligand_entities: list[BoltzGenLigandEntity] = []
    file_entities: list[BoltzGenFileEntity] = []
    constraints: list[BoltzGenConstraint] = []


class BoltzGenScores(Base):
    """
    Compact, interpretable metrics for a designed binder.
    ↑ higher is better, ↓ lower is better

    :param quality_score: aggregate model quality (↑)
    :param num_filters_passed: number of QC/heuristic filters passed (↑)
    :param iptm: inter-chain pTM confidence, 0–1 (↑)
    :param design_ptm: design pTM confidence, 0–1 (↑)
    :param min_interaction_pae: minimum interface PAE in Å (↓)
    :param bb_rmsd: backbone RMSD in Å (↓)
    :param delta_sasa_refolded: ΔSASA of interface after refolding, Å² (↑ typically indicates better burial)
    :param plip_hbonds_refolded: count of hydrogen bonds at the interface (↑)
    :param plip_saltbridge_refolded: count of salt bridges at the interface (↑)
    :param liability_score: composite liabilities score (↓)
    :param liability_high_severity_violations: count of high-severity liabilities (↓)
    :param liability_num_violations: total liability count (↓)
    :param helix: fraction helical content, 0–1
    :param sheet: fraction β-sheet content, 0–1
    :param loop: fraction loop/coil content, 0–1
    :param design_largest_hydrophobic_patch_refolded: largest hydrophobic patch area after refolding, Å²
    :param design_hydrophobicity: overall design hydrophobicity score (unitless)
    :param num_tokens: sequence length / token count
    """

    quality_score: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    num_filters_passed: int | None = None

    iptm: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_ptm: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    min_design_to_target_pae: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_to_target_iptm: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    min_interaction_pae: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    bb_rmsd: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    delta_sasa_refolded: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    plip_hbonds_refolded: int | None = None
    plip_saltbridge_refolded: int | None = None

    liability_score: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    liability_high_severity_violations: int | None = None
    liability_num_violations: int | None = None

    helix: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    sheet: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    loop: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_largest_hydrophobic_patch_refolded: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    design_hydrophobicity: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    num_tokens: int | None = None


class ProteinBinderDesignResult(Base):
    """
    The output; a designed binder.

    :param sequence: the sequence
    :param bound_structure: the PDB of the structure bound to the target
    :param scores: the scores for the generated structure
    """

    binder_sequence: str | None = None  # deprecated
    binder_sequences: list[ProteinSequence] | None = None
    bound_structure: ProteinUUID | None = None
    scores: BoltzGenScores | None = None


class BoltzGenProtocol(LowercaseStrEnum):
    """
    The predefined protocol used for generation + filtering.
    """

    PROTEIN_ANYTHING = "protein-anything"
    PEPTIDE_ANYTHING = "peptide-anything"
    PROTEIN_SMALL_MOLECULE = "protein-small_molecule"
    NANOBODY_ANYTHING = "nanobody-anything"


class BoltzGenSettings(Base):
    """
    The settings for running BoltzGen.

    :param num_designs: how many designs to generate
    :param protocol: which protocol to use
    :param binding_residue: a dict mapping the chain ID to which residues should bind.
        the string follows the BoltzGen format of specifying ranges of residue indices (refer to their documentation).
        examples include "5..7,13" or "5..15,50..".
    """

    protocol: BoltzGenProtocol = BoltzGenProtocol.PROTEIN_ANYTHING
    num_designs: int = 100
    budget: int = 20


class ProteinBinderDesignWorkflow(Workflow):
    """
    A workflow for generating proteins or peptides that bind to something.


    New:
    :param binder_design_input: the input to the protein binder design workflow
    :param binder_design_settings: the settings for the protein generation method employed
    :param generated_binders: the output structures
    """

    binder_design_input: BoltzGenInput = BoltzGenInput()
    binder_design_settings: BoltzGenSettings = BoltzGenSettings()

    generated_binders: list[ProteinBinderDesignResult] = []



from typing import Annotated

from pydantic import AfterValidator

from ..base import Base, round_float
from ..conformers import ConformerClusteringSettings, ConformerProperties, ConformerSearchMixin, iMTDSettings
from ..mode import Mode
from ..settings import Settings
from ..solvent import Solvent, SolventModel, SolventSettings
from ..types import UUID
from .multistage_opt import MultiStageOptSettings
from .workflow import MoleculeWorkflow


class SolventDependentConformer(Base):
    """
    Stores a single conformer scored in many different solvents.

    :param calculation: the conformer (as a calculation)
    :param free_energy_by_solvent: the free energy of the solvent in every solvent (in Hartree)
    :param relative_free_energy_by_solvent: the relative free energy vs. the lowest-energy conformer in every solvent (in kcal/mol)
    :param population_by_solvent: the population of the conformer in every solvent (as a number between 0 and 1)
    """

    calculation: UUID
    free_energy_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]]
    relative_free_energy_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]]
    population_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]]


class SolventDependentConformersWorkflow(ConformerSearchMixin, MoleculeWorkflow):
    """
    Conformers are generated through `conf_gen_settings`.

    Clustering is then performed, and optimization is conducted.
    The optimization of the conformers is governed by `multistage_opt_settings`.

    Final conformer scoring is done through a multi-level scheme:
        - A single-point gas-phase energy run through `multistage_opt_settings.sp_settings`.
        - A separate thermal free-energy correction computed via a single-point Hessian (GFN2-xTB).
        - A per-solvent CPCM-X calculation.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param conf_gen_settings: settings for conformer generation
    :param multistage_opt_settings: set by mode unless mode=MANUAL (ignores additional settings if set)

    New:
    :param solvents: the solvents to study
    :param conformers: output conformers with per-solvent energies and weights
    :param per_solvent_properties: metrics for how overall distribution changes by solvent
    :param relative_free_energy_by_solvent: how free energy changes by solvent, for predicting ∆G_transfer
    """

    solvents: list[Solvent] = [
        Solvent.HEXANE,
        Solvent.OCTANOL,
        Solvent.CHLOROFORM,
        Solvent.DIMETHYLSULFOXIDE,
        Solvent.WATER,
    ]

    conf_gen_settings: iMTDSettings = iMTDSettings(
        max_confs=None,
        mode=Mode.MANUAL,
        speed="normal",
        solvent_settings=SolventSettings(solvent=Solvent.WATER, model=SolventModel.ALPB),
        reopt=False,
    )

    conformer_clustering_settings: ConformerClusteringSettings | None = ConformerClusteringSettings()

    multistage_opt_settings: MultiStageOptSettings = MultiStageOptSettings(
        optimization_settings=[
            Settings(
                method="gfn2_xtb",
                solvent_settings=SolventSettings(
                    solvent=Solvent.WATER,
                    model=SolventModel.ALPB,
                ),
                tasks=["optimize"],
            )
        ],
        sp_settings=Settings(method="g_xtb", tasks=["energy"]),
        mode=Mode.MANUAL,
    )

    conformers: list[SolventDependentConformer] = []
    per_solvent_properties: dict[Solvent, ConformerProperties] = {}
    relative_free_energy_by_solvent: dict[Solvent, Annotated[float, AfterValidator(round_float(3))]] = {}



from abc import ABC
from typing import Annotated, Literal, Self, Sequence

from pydantic import AfterValidator, BaseModel, Field, NonNegativeFloat, PositiveFloat, PositiveInt, field_validator, model_validator

from .base import LowercaseStrEnum, round_float
from .constraint import Constraint
from .method import Method, XTBMethod
from .mode import Mode
from .settings import Settings
from .solvent import SolventModel, SolventSettings
from .workflows.multistage_opt import MultiStageOptMixin


class ConformerProperties(BaseModel):
    """
    Descriptors of a conformer's properties.

    :param solvent_accessible_surface_area: the average SASA, (Å²)
    :param polar_solvent_accessible_surface_area: the average SASA for non-C/H elements (Å²)
    :param radius_of_gyration: the radius of gyration (Å)
    """

    solvent_accessible_surface_area: Annotated[PositiveFloat, AfterValidator(round_float(3))]
    polar_solvent_accessible_surface_area: Annotated[NonNegativeFloat, AfterValidator(round_float(3))]
    radius_of_gyration: Annotated[PositiveFloat, AfterValidator(round_float(3))]


class ConformerClusteringDescriptor(LowercaseStrEnum):
    """Potential descriptors to employ in conformer clustering."""

    SOLVENT_ACCESSIBLE_SURFACE_AREA = "solvent_accessible_surface_area"
    POLAR_SOLVENT_ACCESSIBLE_SURACE_AREA = "polar_solvent_accessible_surface_area"
    RADIUS_OF_GYRATION = "radius_of_gyration"
    PLANE_OF_BEST_FIT = "plane_of_best_fit"
    NORMALIZED_PRINCIPAL_MOMENT_RATIO_1 = "normalized_principal_moment_ratio_1"
    NORMALIZED_PRINCIPAL_MOMENT_RATIO_2 = "normalized_principal_moment_ratio_2"


class ConformerClusteringSettings(BaseModel):
    """
    Settings for clustering conformers based on their three-dimensional properties.

    The properties used for clustering by default are:
    - Solvent-accessible surface area
    - Polar solvent-accessible surface area
    - Radius of gyration
    - Plane of best fit
    - Normalized principal moment ratios 1 and 2

    Rowan uses k-means clustering to identify representative conformers.
    This loosely follows Wilcken and co-workers (10.1007/s10822-020-00337-7).

    :param num_clusters: the number of clusters to include
    :param conformers_per_cluster: the number of compounds to pick from each cluster
    """

    descriptors: list[ConformerClusteringDescriptor] = [
        ConformerClusteringDescriptor.SOLVENT_ACCESSIBLE_SURFACE_AREA,
        ConformerClusteringDescriptor.POLAR_SOLVENT_ACCESSIBLE_SURACE_AREA,
        ConformerClusteringDescriptor.RADIUS_OF_GYRATION,
        ConformerClusteringDescriptor.PLANE_OF_BEST_FIT,
        ConformerClusteringDescriptor.NORMALIZED_PRINCIPAL_MOMENT_RATIO_1,
        ConformerClusteringDescriptor.NORMALIZED_PRINCIPAL_MOMENT_RATIO_2,
    ]

    num_clusters: PositiveInt = 5
    conformers_per_cluster: PositiveInt = 3


class ConformerGenSettings(BaseModel):
    """
    Conformer generation settings.

    Conformers are generated and an initial screening is performed to remove duplicates and high-energy conformers.

    :param screening: post-generation screening settings
    :param constraints: constraints for conformer generation
    :param nci: add a constraining potential for non-covalent interactions
    :param max_confs: maximum number of conformers to keep
    """

    constraints: Sequence[Constraint] = ()
    nci: bool = False
    max_confs: PositiveInt | None = None

    def __str__(self) -> str:
        """Return a string representation of the ConformerGenSettings."""
        return repr(self)

    def __repr__(self) -> str:
        """Return a string representation of the ConformerGenSettings."""
        extra = ""
        if self.constraints:
            extra += f" constraints={len(self.constraints)}"
        if self.nci:
            extra += " nci "
        if self.max_confs:
            extra += f" max_confs={self.max_confs}"

        return f"<{type(self).__name__} {extra}>"


class ETKDGSettings(ConformerGenSettings):
    """
    Settings for ETKDG conformer generation.

    Inherited:
    :param screening: post-generation screening settings
    :param constraints: constraints for conformer generation
    :param nci: add a constraining potential for non-covalent interactions (not supported in ETKDG)
    :param max_confs: maximum number of conformers to keep

    New:
    :param num_initial_confs: number of initial conformers to generate
    :param num_confs_considered: number of conformers to consider for optimization
    :param num_confs_taken: number of final conformers to take
    :param max_mmff_energy: MMFF energy cutoff
    :param max_mmff_iterations: MMFF optimization iterations
    """

    num_initial_confs: int = 300
    num_confs_considered: int = 100
    max_mmff_iterations: int = 500
    max_mmff_energy: float | None = 30
    settings_type: Literal["etkdg"] = "etkdg"

    @field_validator("constraints")
    def check_constraints(cls, constraints: Sequence[Constraint]) -> Sequence[Constraint]:
        if constraints:
            raise ValueError("ETKDG does not support constraints")

        return tuple(constraints)

    @field_validator("nci")
    def check_nci(cls, nci: bool) -> Literal[False]:
        if nci:
            raise ValueError("ETKDG does not support NCI")

        return nci

    @classmethod
    def from_mode(cls, mode: Mode) -> Self:
        match mode:
            case Mode.RECKLESS:
                num_initial_confs = 200
                num_confs_considered = 50
                max_mmff_iterations = 500
                max_mmff_energy = 30
                max_confs = 20
                max_mmff_energy = 20
            case Mode.RAPID:
                num_initial_confs = 300
                num_confs_considered = 100
                max_mmff_iterations = 500
                max_mmff_energy = 30
                max_confs = 50
                max_mmff_energy = 30
            case _:
                raise NotImplementedError(f"Unsupported mode: {mode}")

        return cls(
            num_initial_confs=num_initial_confs,
            num_confs_considered=num_confs_considered,
            max_mmff_iterations=max_mmff_iterations,
            max_mmff_energy=max_mmff_energy,
            max_confs=max_confs,
        )


class iMTDSpeeds(LowercaseStrEnum):
    MEGAQUICK = "megaquick"
    SUPERQUICK = "superquick"
    QUICK = "quick"
    NORMAL = "normal"
    EXTENSIVE = "extensive"


class iMTDSettings(ConformerGenSettings, ABC):
    """
    Settings for iMTD style conformer generation.

    See https://github.com/crest-lab/crest/blob/5ca82feb2ec4df30a0129db957163c934f085952/src/choose_settings.f90#L202
    and https://github.com/crest-lab/crest/blob/5ca82feb2ec4df30a0129db957163c934f085952/src/confparse.f90#L825
    for how quick, superquick, and megaquick are defined.

    See build_imtd_setings(mode) for sensible defaults.

    Inherited:
    :param screening: post-generation screening settings (not used)
    :param constraints: constraints to add
    :param nci: add an ellipsoide potential around the input structure
    :param max_confs: maximum number of conformers to keep

    New:
    :param mtd_method: method for the metadynamics
    :param mtd_runtype: the algorithm used
    :param speed: speed of the calculations (CREST specific setting)
    :param reopt: re-optimize conformers (corrects for the lack of rotamer metadynamics and GC)
    :param free_energy_weights: calculate frequencies and re-weight based on free energies
    :param energy_window: the energy window used, in kcal/mol (CREST specific setting). if set, overrides the default from the speed.
    :param solvent_settings: the solvent to use, if any
    """

    settings_type: Literal["imtd"] = "imtd"

    mtd_method: XTBMethod = Method.GFN_FF
    mtd_runtype: str = "imtd-gc"
    speed: iMTDSpeeds = iMTDSpeeds.QUICK
    reopt: bool = False
    free_energy_weights: bool = False
    energy_window: Annotated[PositiveFloat, AfterValidator(round_float(1))] | None = None
    solvent_settings: SolventSettings | None = None

    @model_validator(mode="after")
    def validate_and_build_imtdgc_settings(self) -> Self:
        if self.solvent_settings and self.solvent_settings.model not in {SolventModel.ALPB, SolventModel.GBSA}:
            raise ValueError("Only ALPB or GBSA solvent models supported for iMTD conformer search!")

        return self

    @classmethod
    def from_mode(cls, mode: Mode) -> Self:
        """
        Settings for iMTD style conformer generation.

        RECKLESS:
            - GFN-FF//MTD(GFN-FF)
            - Megaquick
                - No GC
                - No rotamer metadynamics
                - Energy window = 5.0
                - Run scaling factor = 0.5
                - 6 MTD runs
        RAPID:
            - GFN0//MTD(GFN-FF)
            - Superquick
                - No GC
                - No rotamer metadynamics
                - Energy window = 5.0
                - Run scaling factor = 0.5
                - 6 MTD runs
        CAREFUL:
            - GFN2//MTD(GFN-FF)
            - Quick
                - GC (for iMTD-GC)
                - Rotamer metadynamics (for iMTD-GC)
                - Energy window = 5.0
                - Run scaling factor = 0.5
                - 6 MTD runs
        METICULOUS:
            - GFN2//MTD(GFN2)
            - "Normal"
                - GC (for iMTD-GC)
                - Rotamer metadynamics (for iMTD-GC)
                - Energy window = 6.0
                - Run scaling factor = 1
                - 14 MTD runs (2 with extreme values)
        """
        match mode:
            case Mode.RECKLESS:  # GFN-FF//MTD(GFN-FF)
                mtd_method = Method.GFN_FF
                speed = iMTDSpeeds.MEGAQUICK
                reopt = True
                max_confs: int | None = 20
            case Mode.RAPID:  # GFN0//MTD(GFN-FF)
                mtd_method = Method.GFN_FF
                speed = iMTDSpeeds.SUPERQUICK
                reopt = True
                max_confs = 50
            case Mode.CAREFUL:  # GFN2//MTD(GFN-FF)
                mtd_method = Method.GFN_FF
                speed = iMTDSpeeds.QUICK
                reopt = False
                max_confs = None
            case Mode.METICULOUS:  # GFN2//MTD(GFN2)
                mtd_method = Method.GFN2_XTB
                speed = iMTDSpeeds.NORMAL
                reopt = False
                max_confs = None
            case _:
                raise NotImplementedError(f"Unsupported mode: {mode}")

        return cls(
            mtd_method=mtd_method,
            speed=speed,
            reopt=reopt,
            max_confs=max_confs,
        )


class iMTDGCSettings(iMTDSettings):
    run_type: str = "imtdgc"


class iMTDsMTDSettings(iMTDSettings):
    run_type: str = "imtd-smtd"


class LyrebirdSettings(ConformerGenSettings):
    """
    Settings for Lyrebird-based conformer generation.

    Inherited:
    :param screening: post-generation screening settings
    :param constraints: constraints for conformer generation (not supported)
    :param nci: add a constraining potential for non-covalent interactions (not supported)
    :param max_confs: maximum number of conformers to keep

    New:
    :param num_initial_confs: number of initial conformers to generate
    """

    num_initial_confs: int = 300
    settings_type: Literal["lyrebird"] = "lyrebird"

    @field_validator("constraints")
    def check_constraints(cls, constraints: Sequence[Constraint]) -> Sequence[Constraint]:
        if constraints:
            raise ValueError("Lyrebird does not support constraints")

        return tuple(constraints)

    @field_validator("nci")
    def check_nci(cls, nci: bool) -> Literal[False]:
        if nci:
            raise ValueError("Lyrebird does not support NCI")

        return nci


class MonteCarloMultipleMinimumSettings(ConformerGenSettings):
    """
    Settings for Monte-Carlo-multiple-minimum-based conformer generation.
    Default values recommended by Nick Casetti.

    Inherited:
    :param screening: post-generation screening settings
    :param constraints: constraints for conformer generation (not supported)
    :param nci: add a constraining potential for non-covalent interactions (not supported)
    :param max_confs: maximum number of conformers to keep

    New:
    :param num_monte_carlo_iterations: number of Monte Carlo iterations to run
    :param rmsd_threshold: the threshold to determine if MCMM output structures are identical
    :param energy_window: maximum energy window above the minimum-energy conformer above which to retain (kcal/mol)
    :param monte_carlo_settings: the way the actual energy will be computed for the Monte-Carlo steps
    """

    energy_settings: Settings = Settings(method=Method.AIMNET2_WB97MD3)

    num_monte_carlo_iterations: int = 250
    rmsd_threshold: float = 0.5
    energy_window: float = 20

    settings_type: Literal["monte_carlo_multiple_minimum"] = "monte_carlo_multiple_minimum"

    @field_validator("constraints")
    def check_constraints(cls, constraints: Sequence[Constraint]) -> Sequence[Constraint]:
        if constraints:
            raise ValueError("MCMM does not support constraints")

        return tuple(constraints)

    @field_validator("nci")
    def check_nci(cls, nci: bool) -> Literal[False]:
        if nci:
            raise ValueError("MCMM does not support NCI")

        return nci


ConformerGenSettingsUnion = Annotated[ETKDGSettings | iMTDSettings | LyrebirdSettings | MonteCarloMultipleMinimumSettings, Field(discriminator="settings_type")]


class ConformerGenMixin(BaseModel):
    """
    Mixin for classes need conformer generation.

    :param conf_gen_settings: settings for conformer generation
    :param constraints: constraints to add
    :param nci: add a constraining potential for non-covalent interactions
    :param max_confs: maximum number of conformers to keep
    :param clustering_settings: how to cluster the conformers (if at all)
    """

    conf_gen_settings: ConformerGenSettingsUnion | None
    constraints: Sequence[Constraint] = ()
    nci: bool = False
    max_confs: int | None = None

    conformer_clustering_settings: ConformerClusteringSettings | None = None


class ConformerSearchMixin(ConformerGenMixin, MultiStageOptMixin):
    """
    Mixin for classes need conformer search—a combination of conformer generation and optimization.

    Inherited (ConformerGenMixin):
    :param conf_gen_settings: settings for conformer generation
    :param nci: add a constraining potential for non-covalent interactions

    Inherited (MultiStageOptMixin):
    :param multistage_opt_settings: settings for the optimization
    :param solvent: solvent to use
    :param xtb_preopt: pre-optimize with xtb
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    Inherited (Both):
    :param constraints: constraints to add (diamond inheritance, works as expected)
    """

    def __str__(self) -> str:
        """Return a string representation of the ConformerSearch workflow."""
        return repr(self)

    @model_validator(mode="after")
    def remove_ts_constraints(self) -> Self:
        """Remove constraints from optimization if a TS."""
        msos = self.multistage_opt_settings
        if msos.transition_state and msos.constraints:
            msos.constraints = []
            for opt_set in msos.optimization_settings:
                opt_set.opt_settings.constraints = []

        return self



from .base import Base, LowercaseStrEnum


class UseSOSCF(LowercaseStrEnum):
    ALWAYS = "always"
    UPON_FAILURE = "upon_failure"
    NEVER = "never"


class SCFSettings(Base):
    """
    Settings for SCF convergence.

    :param max_iters: the maximum number of SCF iterations to permit
    :param soscf: whether or not to use SOSCF (second-order SCF).
    """

    max_iters: int = 250
    soscf: UseSOSCF = UseSOSCF.UPON_FAILURE



from enum import Enum


class Status(int, Enum):
    # what a job gets when it's created if user is below max_concurrency
    QUEUED = 0

    # still running...
    RUNNING = 1

    # what a job gets when it has finished
    COMPLETED_OK = 2

    # something went wrong
    FAILED = 3

    # job was stopped externally, maybe timed out or something
    STOPPED = 4

    # Status if a user has exceeded their allowed max_concurrency
    AWAITING_QUEUE = 5

    # Job is not yet submitted
    DRAFT = 6



"""Ion mobility workflow."""

from typing import Annotated, Self

from pydantic import AfterValidator, computed_field, model_validator

from ..base import Base, round_optional_float
from ..data import ELEMENT_SYMBOL
from ..types import UUID, round_list
from .workflow import MoleculeWorkflow


class IonMobilityForcefieldElement(Base):
    """
    A single atom specification for the ion-mobility forcefield.

    :param name: the name of the element (e.g. "Hydrogen")
    :param atomic_number: the element's atomic number
    :param mass: the mass of the element in Daltons (e.g. 1.00794)
    :param sigma: the sigma Lennard-Jones parameter, in Å
    :param epsilon: the epsilon Lennard-Jones parameter, in kcal/mol
    """

    name: str
    atomic_number: int
    mass: float
    sigma: float
    epsilon: float

    @computed_field  # type: ignore[misc, prop-decorator, unused-ignore]
    @property
    def symbol(self) -> str:
        return ELEMENT_SYMBOL[self.atomic_number]


class IonMobilityWorkflow(MoleculeWorkflow):
    """
    Workflow for calculating hydrogen bond basicity.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param protonate: automatically protonate the molecule
    :param temperature: the temperature, in Kelvin
    :param do_csearch: whether to perform a conformational search
    :param do_optimization: whether to perform an optimization
    :param forcefield: the forcefield used to describe atom–gas interactions.
        if None, the default forcefield will be used.

    Results:
    :param conformers: the UUIDs of the conformers
    :param conformer_ccs: the collision cross section (Å**2) per conformer
    :param conformer_ccs_stdev: the uncertainty in the same
    :param conformer_weights: the Boltzmann weights at RT
    :param average_ccs: the Boltzmann-weighted CCS for the ensemble
    :param average_ccs_stdev: the uncertainty in the same
    """

    protonate: bool = False
    temperature: float = 300
    do_csearch: bool = True
    do_optimization: bool = True

    forcefield: list[IonMobilityForcefieldElement] | None = None

    conformers: list[UUID] = []

    conformer_ccs: Annotated[list[float], AfterValidator(round_list(3))] = []
    conformer_ccs_stdev: Annotated[list[float], AfterValidator(round_list(3))] = []
    boltzmann_weights: Annotated[list[float], AfterValidator(round_list(3))] = []

    average_ccs: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    average_ccs_stdev: Annotated[float | None, AfterValidator(round_optional_float(3))] = None

    @model_validator(mode="after")
    def check_supported_atoms(self) -> Self:
        """Validate that user-supplied forcefields have correct atoms."""
        if self.forcefield is not None:
            supported_atoms = {e.atomic_number for e in self.forcefield}
            if not all(z in supported_atoms for z in self.initial_molecule.atomic_numbers):
                raise ValueError("Provided forcefield does not support all elements in input structure!")

        return self



"""Molecular descriptors workflow."""

from ..types import UUID
from .workflow import MoleculeWorkflow

Descriptors = dict[str, dict[str, float] | tuple[float | None, ...] | float]


class DescriptorsWorkflow(MoleculeWorkflow):
    """
    A workflow for calculating molecular descriptors.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow

    New:
    :param optimization: UUID of optimization
    :param descriptors: calculated descriptors
    """

    optimization: UUID | None = None

    descriptors: Descriptors | None = None



"""Molecular dynamics workflow."""

from typing import Self

from pydantic import PositiveFloat, PositiveInt, computed_field, model_validator

from ..base import Base, LowercaseStrEnum
from ..constraint import PairwiseHarmonicConstraint, SphericalHarmonicConstraint
from ..engine import Engine
from ..settings import Settings
from ..types import UUID
from .workflow import MoleculeWorkflow


class MolecularDynamicsInitialization(LowercaseStrEnum):
    """Initialization method for molecular dynamics."""

    RANDOM = "random"
    QUASICLASSICAL = "quasiclassical"
    READ = "read"


class ThermodynamicEnsemble(LowercaseStrEnum):
    """Thermodynamic ensemble for molecular dynamics."""

    NPT = "npt"
    NVT = "nvt"
    NVE = "nve"


class Frame(Base):
    """A frame in a molecular dynamics simulation."""

    index: int  # what number frame this is within the MD simulation

    calculation_uuid: UUID | None = None  # UUID of calculation

    pressure: float
    temperature: float
    volume: float
    potential_energy: float  # kcal/mol
    kinetic_energy: float  # kcal/mol

    @computed_field  # type: ignore[misc, prop-decorator, unused-ignore]
    @property
    def energy(self) -> float:
        return self.potential_energy + self.kinetic_energy


class MolecularDynamicsSettings(Base):
    """Settings for a molecular dynamics simulation."""

    ensemble: ThermodynamicEnsemble = ThermodynamicEnsemble.NVT
    initialization: MolecularDynamicsInitialization = MolecularDynamicsInitialization.RANDOM

    timestep: PositiveFloat = 1.0  # fs
    num_steps: PositiveInt = 500
    save_interval: PositiveInt = 10

    confining_constraint: SphericalHarmonicConstraint | None = None

    temperature: PositiveFloat = 300  # K
    pressure: PositiveFloat | None = None  # atm

    langevin_thermostat_timescale: PositiveFloat = 100  # fs
    berendsen_barostat_timescale: PositiveFloat = 1000  # fs

    constraints: list[PairwiseHarmonicConstraint] = []

    @model_validator(mode="after")
    def validate_ensemble_settings(self) -> Self:
        """Check that NVT ensemble always has temperature defined, and that NPT has temp and pressure defined."""
        if self.ensemble == ThermodynamicEnsemble.NVT and self.temperature is None:
            raise ValueError("NVT ensemble must have a temperature defined")
        elif self.ensemble == ThermodynamicEnsemble.NPT and (self.temperature is None or self.pressure is None):
            raise ValueError("NPT ensemble must have both temperature and pressure defined")
        return self


class MolecularDynamicsWorkflow(MoleculeWorkflow):
    """
    Workflow for running a molecular dynamics simulation.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param mode: Mode for workflow (currently unused)

    New:
    :param settings: settings for the molecular dynamics simulation
    :param calc_settings: settings for the gradient calculation
    :param calc_engine: engine to use for the gradient calculation [uses calc_settings.method.default_engine() if not set]

    Results:
    :param frames: Frames from the MD
    """

    settings: MolecularDynamicsSettings
    calc_settings: Settings

    # DEPRECATED - will remove later
    calc_engine: Engine = None  # type: ignore[assignment]

    frames: list[Frame] = []

    @model_validator(mode="after")
    def validate_calc_engine(self) -> Self:
        """Ensure that the calc_engine is set."""
        self.calc_engine = self.calc_engine or self.calc_settings.engine

        return self



"""Settings for the Freezing String Method (FSM)."""

from typing import Annotated, Self

from pydantic import AfterValidator, BaseModel, PositiveFloat, PositiveInt, model_validator

from ..base import LowercaseStrEnum, round_float


class FSMOptimizationCoordinates(LowercaseStrEnum):
    """Coordinate systems for FSM optimization step."""

    CARTESIAN = "cartesian"
    REDUNDANT_INTERNAL_COORDINATES = "redundant_internal_coordinates"


class FSMInterpolation(LowercaseStrEnum):
    """Methods for FSM interpolation step."""

    CARTESIAN = "cartesian"
    LINEAR_SYNCHRONOUS_TRANSIT = "linear_synchronous_transit"
    REDUNDANT_INTERNAL_COORDINATES = "redundant_internal_coordinates"


class FSMSettings(BaseModel):
    """
    Settings for the Freezing String Method (FSM) TS search.

    :param opt_coords: coordinate system to use for optimization
    :param interpolation_method: method to use for interpolation between nodes
    :param min_num_nodes: minimum number of nodes to use in the string
    :param num_interpolation_points: number of interpolation points to use between end nodes
    :param max_optimizer_iterations: maximum number of optimizer iterations to perform  (scipy.minimize maxiter)
    :param max_line_search_steps: maximum number of line search steps to perform (scipy.minimize maxls)
    :param max_displacement: maximum displacement for a single coordinate (Å for distances, internally converted for angles)
    """

    optimization_coordinates: FSMOptimizationCoordinates = FSMOptimizationCoordinates.CARTESIAN
    interpolation_method: FSMInterpolation = FSMInterpolation.REDUNDANT_INTERNAL_COORDINATES

    min_num_nodes: PositiveInt = 18
    num_interpolation_points: PositiveInt = 100

    max_optimizer_iterations: PositiveInt = 3
    max_line_search_steps: PositiveInt = 2
    max_displacement: Annotated[PositiveFloat, AfterValidator(round_float(3))] = 0.3

    @model_validator(mode="after")
    def sanity_check(self) -> Self:
        """Sanity check for settings."""
        if self.min_num_nodes < 5:
            raise ValueError("Must have at least 5 nodes for the Freezing String Method.")
        if self.max_displacement < 0.001:
            raise ValueError("Maximum displacement must be at least 0.001 Å.")

        return self



from abc import ABC
from typing import Annotated, Literal, Self

from pydantic import BaseModel, Field, PositiveInt, model_validator


class ExcitedStateSettings(BaseModel, ABC):
    """Settings for excited state calculations."""


class TDDFTSettings(ExcitedStateSettings):
    """
    Settings for TDDFT calculations.

    New:
    :param tda: use Tamm-Dancoff approximation
    :param num_excitations: number of excitations to calculate
    :param target_root: root to target (for gradient/optimization)
    """

    tda: bool = True
    num_excitations: PositiveInt = 5
    target_root: PositiveInt | None = None

    settings_type: Literal["tddft_settings"] = "tddft_settings"

    @model_validator(mode="after")
    def validate_setup(self) -> Self:
        if self.target_root and (self.target_root > self.num_excitations):
            raise ValueError("target_root cannot be greater than num_excitations.")

        return self


ExcitedStateSettingsUnion = Annotated[TDDFTSettings, Field(discriminator="settings_type")]



"""Protein cofolding workflow."""

from typing import Annotated, Literal, TypeAlias

from pydantic import AfterValidator, BaseModel, ConfigDict

from ..base import Base, LowercaseStrEnum, round_float, round_optional_float
from ..types import UUID, round_list
from .workflow import FASTAWorkflow

ProteinUUID: TypeAlias = UUID
CalculationUUID: TypeAlias = UUID


class CofoldingModel(LowercaseStrEnum):
    """Cofolding model to be used for prediction."""

    CHAI_1R = "chai_1r"
    BOLTZ_1 = "boltz_1"
    BOLTZ_2 = "boltz_2"


class Token(BaseModel):
    """Either an atom in a ligand or a residue in a protein or nucleotide chain."""

    input_type: Literal["ligand", "protein", "dna", "rna"]
    input_index: int
    token_index: int
    atom_name: str | None = None


class ContactConstraint(BaseModel):
    """Contact constraint to be used for prediction."""

    token_1: Token
    token_2: Token
    max_distance: float  # Angstroms
    force: bool = False  # Whether to use potentials to enforce the constraint


class PocketConstraint(BaseModel):
    """Pocket constraint to be used for prediction."""

    input_type: Literal["ligand", "protein", "dna", "rna"]
    input_index: int
    contacts: list[Token]
    max_distance: float  # Angstroms
    force: bool = False  # Whether to use potentials to enforce the constraint


class CofoldingScores(BaseModel):
    """The output scores from co-folding scores."""

    confidence_score: Annotated[float, AfterValidator(round_float(3))]
    ptm: Annotated[float, AfterValidator(round_float(3))]  # predicted template modeling score
    iptm: Annotated[float, AfterValidator(round_float(3))]  # interface predicted template modeling score
    avg_lddt: Annotated[float, AfterValidator(round_float(3))]


class AffinityScore(BaseModel):
    pred_value: Annotated[float, AfterValidator(round_float(3))]
    probability_binary: Annotated[float, AfterValidator(round_float(3))]
    pred_value1: Annotated[float, AfterValidator(round_float(3))]
    probability_binary1: Annotated[float, AfterValidator(round_float(3))]
    pred_value2: Annotated[float, AfterValidator(round_float(3))]
    probability_binary2: Annotated[float, AfterValidator(round_float(3))]


class CofoldingResult(Base):
    """Results for a single cofolding sample.

    :param affinity_score: affinity score
    :param lddt: local distance different test result
    :param predicted_structure_uuid: UUID of the predicted structure
    :param scores: output cofolding scores
    :param pose: UUID of the calculation pose
    :param strain: strain of the ligand, in kcal/mol
    :param predicted_refined_structure_uuid: if the structure has been refined,
        UUID of the predicted structure after refinement
    """

    lddt: Annotated[list[float] | None, AfterValidator(round_list(3))] = None
    affinity_score: AffinityScore | None = None
    predicted_structure_uuid: ProteinUUID | None = None
    scores: CofoldingScores | None = None
    pose: CalculationUUID | None = None
    posebusters_valid: bool | None = None
    strain: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    predicted_refined_structure_uuid: ProteinUUID | None = None


class ProteinCofoldingWorkflow(FASTAWorkflow):
    """
    Workflow for predicting structures.

    Especially protein structures. At least one biological sequence is required.

    Inherited:
    :param initial_protein_sequences: protein sequences of interest
    :param initial_dna_sequences: DNA sequences of interest
    :param initial_rna_sequences: RNA sequences of interest
    :param initial_smiles_list: SMILES strings of interest

    New:
    :param use_msa_server: whether to use the MSA server
    :param use_templates_server: whether to use the templates server
    :param use_potentials: whether to use the potentials (inference-time steering) with Boltz
    :param contact_constraints: Boltz contact constraints
    :param pocket_constraints: Boltz pocket constraints
    :param do_pose_refinement: whether to optimize non-rotatable bonds in output poses
    :param compute_strain: whether to compute the strain of the pose (if `pose_refinement` is enabled)
    :param num_samples: number of samples generated for prediction
    :param model: which cofolding model to use
    :param cofolding_results: per diffusion sample outputs, grouped together
    """

    model_config = ConfigDict(validate_assignment=True)

    use_msa_server: bool = False
    use_templates_server: bool = False
    use_potentials: bool = False
    contact_constraints: list[ContactConstraint] = []
    pocket_constraints: list[PocketConstraint] = []
    do_pose_refinement: bool = False
    compute_strain: bool = False
    num_samples: int | None = None

    model: CofoldingModel = CofoldingModel.BOLTZ_2
    cofolding_results: list[CofoldingResult] | None = None

    affinity_score: AffinityScore | None = None
    lddt: Annotated[list[float] | None, AfterValidator(round_list(3))] = None
    predicted_structure_uuid: ProteinUUID | None = None
    scores: CofoldingScores | None = None
    pose: CalculationUUID | None = None
    posebusters_valid: bool | None = None
    strain: Annotated[float | None, AfterValidator(round_optional_float(3))] = None
    predicted_refined_structure_uuid: ProteinUUID | None = None



"""Conformer search workflow."""

from collections import Counter
from typing import Annotated, Self

from pydantic import AfterValidator, Field, model_validator

from ..conformers import ConformerProperties, ConformerSearchMixin, iMTDSettings
from ..molecule import Molecule
from ..types import UUID, FloatPerAtom, round_float_per_atom
from .workflow import MoleculeWorkflow, SMILESWorkflow


class ConformerSearchWorkflow(ConformerSearchMixin, SMILESWorkflow, MoleculeWorkflow):
    """
    ConformerSearch Workflow.

    This workflow supports both SMILES and 3D molecular input. Some conformer generation settings
    support both methods; others (like CREST) require 3D information. Only one should be supplied.

    Inherited:
    :param initial_molecule: Molecule of interest
    :param initial_smiles: SMILES of the molecule of interest
    :param conf_gen_settings: settings for conformer generation
    :param multistage_opt_settings: multi-stage optimization settings
    :param solvent: solvent to use
    :param xtb_preopt: pre-optimize with xtb
    :param constraints: constraints to add
    :param transition_state: whether this is a transition state
    :param frequencies: whether to calculate frequencies

    New:
    :param initial_conformers: input conformers (if no conformer-generation is requested)
    :param conformer_uuids: list of UUIDs of the Molecules generated
    :param energies: energies of the molecules
    :param conformer_properties: each conformer's properties
    :param ensemble_properties: the overall ensemble's properties
    """

    initial_smiles: str = ""
    initial_molecule: Molecule | None = None  # type: ignore [assignment]
    initial_conformers: list[Molecule] = []

    # Results
    conformer_uuids: list[list[UUID | None]] = Field(default_factory=list)
    energies: Annotated[FloatPerAtom, AfterValidator(round_float_per_atom(6))] = Field(default_factory=list)

    conformer_properties: list[ConformerProperties] = []
    ensemble_properties: ConformerProperties | None = None

    @model_validator(mode="after")
    def validate_mol_input(self) -> Self:
        """Ensure that a valid combination of input types is set."""
        if (not self.initial_conformers) and (not (bool(self.initial_smiles) ^ bool(self.initial_molecule))):
            raise ValueError("Can only set one of initial_molecule and initial_smiles")

        if isinstance(self.conf_gen_settings, iMTDSettings) and (self.initial_molecule is None):
            raise ValueError("iMTDSettings requires initial_molecule to be set")

        if self.conf_gen_settings is None and not self.initial_conformers:
            raise ValueError("Need initial_conformers to be set without a conformer-generation method")

        if len(self.initial_conformers) > 1:
            initial_count = Counter(self.initial_conformers[0].atomic_numbers)
            for conformer in self.initial_conformers[1:]:
                if Counter(conformer.atomic_numbers) != initial_count:
                    raise ValueError("Not all molecules in initial_conformers have the same atomic formula")

        return self
